



sed -i '/brotlipy/d' tests/requirements-py2.txt
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip install setuptools-scm
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip install -ctests/constraints.txt -rrequirements-py3.txt -rtests/requirements-py3.txt Pillow botocore google-cloud-storage
IF %ERRORLEVEL% NEQ 0 exit /B 1
mv scrapy _scrapy
IF %ERRORLEVEL% NEQ 0 exit /B 1
rm tests/test_command_parse.py tests/test_command_shell.py tests/test_commands.py tests/test_crawler.py tests/test_downloader_handlers.py tests/test_exporters.py tests/test_feedexport.py tests/test_proxy_connect.py tests/test_utils_trackref.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest _scrapy tests
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
